//=========================================================
//  MusE
//  Linux Music Editor
//  $Id: ./muse/fastlog.h $
//
//  Copyright (C) 1999-2011 by Werner Schweer and others
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; version 2 of
//  the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
//=========================================================
/* Copyright unknown. Code by Laurent de Soras <laurent@ohmforce.com>.
 */

#ifndef __pbd_fastlog_h__
#define __pbd_fastlog_h__

#include "muse_math.h" /* Get math.h for HUGE_VAL */

namespace MusECore {

static inline float fast_log2 (float val)
      {
	/* don't use reinterpret_cast<> because that prevents this
	   from being used by pure C code (for example, GnomeCanvasItems)
	*/
      union {float f; int i;} t;
	t.f = val;
      int* const exp_ptr = &t.i;
	int x              = *exp_ptr;
	const int log_2    = ((x >> 23) & 255) - 128;
	x &= ~(255 << 23);
	x += 127 << 23;
	*exp_ptr = x;
	val = ((-1.0f/3) * t.f + 2) * t.f - 2.0f/3;
	return (val + log_2);
      }

static inline float fast_log (const float val)
      {
      return (fast_log2 (val) * 0.69314718f);
      }

static inline float fast_log10 (const float val)
      {
	return fast_log2(val) / 3.312500f;
      }

static inline float minus_infinity() { return -HUGE_VAL; }

} // namespace MusECore

#endif /* __pbd_fastlog_h__ */
